﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=StaticCode
Version=6.01
@EndOfDesignText@
'Static code module
Sub Process_Globals
	Private fx As JFX
End Sub

Public Sub ApplySha256(Input As String) As String
	
	Dim Digest As MessageDigest
	Dim Hash() As Byte = Digest.GetMessageDigest(Input.GetBytes("UTF-8"),"SHA-256")
	
	Dim HexString As StringBuilder
	HexString.Initialize
	
	Dim Integer As JavaObject
	Integer.InitializeStatic("java.lang.Integer")
	
	For i = 0 To Hash.Length - 1
		Dim Hex As String = Integer.RunMethod("toHexString",Array(Bit.And(Hash(i),0xFF)))
		If Hex.Length = 1 Then HexString.Append("0")
		HexString.Append(Hex)
	Next
	
	Return HexString.ToString
End Sub

Public Sub IsChainValid(Chain As List, Difficulty As Int, GenesisTransaction As Transaction) As Boolean
	Dim CurrentBlock As Block
	Dim PreviousBlock As Block
	
	Dim HashTarget As String = GetDifficultyString(Difficulty)
	
	Dim TempUTXOs As Map
	TempUTXOs.Initialize
	Dim GTO As TransactionOutput = GenesisTransaction.Outputs.Get(0)
	TempUTXOs.Put(GTO.Id,GTO)
	
	
	For i = 1 To Chain.Size - 1
		CurrentBlock = Chain.Get(i)
		PreviousBlock = Chain.Get(i-1)
		
		'compare registered hash and calculated hash:
		If Not(CurrentBlock.Hash.EqualsIgnoreCase(CurrentBlock.CalculateHash)) Then
			Log("Current Hashes Not Equal")
			Return False
		End If
		
		'compare previous hash and registered previous hash
		If Not(PreviousBlock.Hash.EqualsIgnoreCase(CurrentBlock.PreviousHash)) Then
			Log("Previous Hashes Not Equal")
			Return False
		End If
		
		'check if hash is solved
		If Not(CurrentBlock.Hash.SubString2(0,Difficulty).EqualsIgnoreCase(HashTarget)) Then
			Log("This block hasn't been mined")
			Return False
		End If
		
	Next
	
	'Loop through blockchains transactions
	
	Dim TempOutput As TransactionOutput
	
	For t = 0 To CurrentBlock.Transactions.Size - 1
	
		Dim CurrentTransaction As Transaction = CurrentBlock.Transactions.get(t)
		If Not(CurrentTransaction.VerifySignature) Then
			Log($"Signature on Transaction ${t} is invalid"$)
			Return False
		End If
		
		If CurrentTransaction.getInputsValue <> CurrentTransaction.getOutputsValue Then
			Log($"Inputs are note equal to outputs on Transaction ${t})"$)
			Return False
		End If
		
		For Each input As TransactionInput In CurrentTransaction.Inputs
			TempOutput = TempUTXOs.get(input.transactionOutputId)
			
			If TempOutput <> Null Or TempOutput.IsInitialized = False Then
				Log("Referenced input on Transaction(" & t & ") is Missing")
				Return False
			End If
			
			If input.UTXO.value <> TempOutput.value Then
				Log("#Referenced input Transaction(" & t & ") value is Invalid")
				Return False
			End If
					
			TempUTXOs.remove(input.transactionOutputId)
			
		Next
		
		For Each Output As TransactionOutput In CurrentTransaction.Outputs
			TempUTXOs.Put(Output.Id,Output)
		Next
		
		TempOutput = CurrentTransaction.Outputs.Get(0)
		
		If TempOutput.Recipient <> CurrentTransaction.Recipient Then
			Log("Transaction(" & t & ") output reciepient is not who it should be")
			Return False
		End If
		
		TempOutput = CurrentTransaction.Outputs.Get(1)
		
		If TempOutput.Recipient <> CurrentTransaction.Sender_ Then
			Log("Transaction(" & t & ") output 'change' is not to sender")
			Return False
		End If
		
		
	Next
	
	Return True
End Sub

'Applies ECDSA Signature and returns the result ( as bytes ).
Public Sub applyECDSASig(PrivateKey As JavaObject, Input As String) As Byte()
	Dim dsa As JavaObject
	
	Dim Signature As JavaObject
	Signature.InitializeStatic("java.security.Signature")
	
	dsa = Signature.RunMethod("getInstance",Array("ECDSA", "BC"))
	dsa.RunMethod("initSign",Array(PrivateKey))
	Dim StrByte() As Byte = Input.GetBytes("UTF-8")
	dsa.RunMethod("update",Array(StrByte))
	Dim RealSig() As Byte = dsa.RunMethod("sign",Null)
	Return RealSig
	
	
End Sub

Public Sub VerifyECDSASig(PublicKey As JavaObject,Data As String, TSignature() As Byte) As Boolean
	Dim Signature As JavaObject
	Signature.InitializeStatic("java.security.Signature")
	
	Dim EcdsaVerify As JavaObject
	EcdsaVerify = Signature.RunMethod("getInstance",Array("ECDSA", "BC"))
	EcdsaVerify.RunMethod("initVerify",Array(PublicKey))
	EcdsaVerify.RunMethod("update",Array(Data.GetBytes("UTF-8")))
	
	Return EcdsaVerify.RunMethod("verify",Array(TSignature))
End Sub

Public Sub GetStringFromKey(Key As JavaObject) As String
	Dim Base64 As JavaObject
	Base64.InitializeStatic("java.util.Base64")
	Return Base64.RunMethodJO("getEncoder",Null).RunMethod("encodeToString",Array(Key.RunMethod("getEncoded",Null)))
End Sub

Public Sub GetMerkleRoot(Transactions As List) As String
	Dim Count As Int = Transactions.Size
	Dim PreviousTreeLayer As List
	PreviousTreeLayer.Initialize
	
	For Each Trans As Transaction In Transactions
		PreviousTreeLayer.Add(Trans.TransactionID)
	Next
	
	Dim TreeLayer As List
	TreeLayer.Initialize
	TreeLayer.AddAll(PreviousTreeLayer)
	
	Do While Count > 1
		Dim TreeLayer As List
		TreeLayer.Initialize
		For i = 1 To PreviousTreeLayer.Size - 1 Step 2
			TreeLayer.Add(ApplySha256(PreviousTreeLayer.Get(i-1) & PreviousTreeLayer.Get(i)))
		Next
		Count = TreeLayer.Size
		PreviousTreeLayer.Clear
		PreviousTreeLayer.AddAll(TreeLayer)
	Loop
	
	If TreeLayer.Size = 1 Then
		Return TreeLayer.Get(0)
	Else
		Return 0
	End If
End Sub

Public Sub GetDifficultyString(Difficulty As Int) As String
	Dim SB As StringBuilder
	SB.Initialize
	For i = 0 To Difficulty - 1
		SB.Append("0")
	Next
	Return SB.ToString
End Sub
