﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=6.01
@EndOfDesignText@
Sub Class_Globals
	Private fx As JFX
	
	Public TransactionID As String 		'this is also the hash of the transaction.
	Public Sender_ As JavaObject 		'senders address/public key.
	Public Recipient As JavaObject 	'Recipients address/public key.
	Public Value As Double
	Public Signature() As Byte			'this is to prevent anybody else from spending funds in our wallet.
	
	Public Inputs As List
	Public Outputs As List
	
	Private Sequence As Int = 0
End Sub

'Initializes the object. You can add parameters to this method if needed.
Public Sub Initialize(From As JavaObject, TRecipient As JavaObject, TValue As Double, TInputs As List)
	Sender_ = From
	Recipient = TRecipient
	Value = TValue
	Inputs = TInputs
	Outputs.Initialize
End Sub

Public Sub CalculateHash As String
	Sequence = Sequence + 1			'increase the sequence to avoid 2 identical transactions having the same hash
	Return BCUtils.ApplySha256(BCUtils.getStringFromKey(Sender_) & BCUtils.getStringFromKey(Recipient) & Value & Sequence)
End Sub

Public Sub GenerateSignature(PrivateKey As JavaObject)
	Dim Data As String = BCUtils.GetStringFromKey(Sender_) & BCUtils.GetStringFromKey(Recipient) & Value
	Signature = BCUtils.applyECDSASig(PrivateKey,Data)
End Sub

Public Sub VerifySignature As Boolean
	Dim Data As String = BCUtils.GetStringFromKey(Sender_) & BCUtils.GetStringFromKey(Recipient) & Value
	Return BCUtils.VerifyECDSASig(Sender_,Data,Signature)
End Sub

Public Sub ProcessTransaction As Boolean
	
	If VerifySignature = False Then 
		Log("Transaction Signature failed to verify")
		Return False
	End If
	
	For Each I As TransactionInput In Inputs
		I.UTXO = Main.UTXOs.Get(I.TransactionOutputID)
	Next
	
	'Check If Transaction is Valid
	If GetInputsValue < Main.MinimumTransaction Then
		Log("Transaction Inputs too small")
		Return False
	End If
	
	Dim Leftover As Double = GetInputsValue - Value 'get value of inputs then the left over change:
	
	TransactionID = CalculateHash
	
	'!!!!!!!!!!!!!!!!!!!!!!!!!!! Both transactions will have the same Hash, not yet sure if this is a bug!!
	
	Dim TOut As TransactionOutput
	TOut.Initialize(Recipient,Value,TransactionID)    'send value to recipient
	Outputs.Add(TOut)
	
	Dim TOut As TransactionOutput
	TOut.Initialize(Sender_,Leftover,TransactionID)    'send the left over 'change' back to sender
	Outputs.Add(TOut)
	
	'add outputs to Unspent list
	For Each O As TransactionOutput In Outputs
		Main.UTXOs.Put(O.Id, O)
	Next
	
	'remove transaction inputs from UTXO lists as spent:
	
	For Each i As TransactionInput In Inputs
		Main.UTXOs.Remove(I.UTXO.Id)
	Next
	
	
	Return True
End Sub

'returns sum of inputs(UTXOs) values
Public Sub GetInputsValue As Double
	Dim Total As Double = 0
	For Each i As TransactionInput In Inputs
		If i.UTXO.IsInitialized = False Then Continue
		Total = Total + i.UTXO.Value
	Next
	
	Return Total
End Sub

'returns sum of outputs:
Public Sub GetOutputsValue As Double
	Dim Total As Double = 0
	For Each o As TransactionOutput In Outputs
		Total = Total + o.Value
	Next
	
	Return Total
End Sub
